@@ -189,7 +189,7 @@ public abstract class BaseObjectPoolConfig implements Cloneable {
     // TODO Consider changing this to a single property for 3.x
     private String jmxNamePrefix = DEFAULT_JMX_NAME_PREFIX;
 
-    private String jmxNameBase = DEFAULT_JMX_NAME_PREFIX;
+    private String jmxNameBase = DEFAULT_JMX_NAME_BASE;
 
 
     /**
