@@ -29,6 +29,7 @@ import java.io.PrintWriter;
 import java.io.StringWriter;
 import java.lang.management.ManagementFactory;
 import java.util.ArrayList;
+import java.util.HashSet;
 import java.util.NoSuchElementException;
 import java.util.Random;
 import java.util.Set;
@@ -62,7 +63,7 @@ import org.junit.Before;
 import org.junit.Test;
 
 /**
- * @version $Revision: 1648001 $
+ * @version $Revision: 1679448 $
  */
 public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
 
@@ -1848,6 +1849,7 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         }
         @Override
         public boolean validateObject(K key, PooledObject<String> obj) {
+            doWait(validateLatency);
             if (enableValidation) {
                 return validateCounter++%2 == 0 ? evenValid : oddValid;
             } else {
@@ -1878,6 +1880,9 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         public void setMakeLatency(long makeLatency) {
             this.makeLatency = makeLatency;
         }
+        public void setValidateLatency(long validateLatency) {
+            this.validateLatency = validateLatency;
+        }
         public void setValidationEnabled(boolean b) {
             enableValidation = b;
         }
@@ -1911,6 +1916,7 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         boolean enableValidation = false;
         long destroyLatency = 0;
         long makeLatency = 0;
+        long validateLatency = 0;
         volatile int maxTotalPerKey = Integer.MAX_VALUE;
         boolean exceptionOnPassivate = false;
         boolean exceptionOnActivate = false;
@@ -2136,6 +2142,110 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         factory.exceptionOnCreate = false;
         pool.borrowObject("One");
     }
+    
+    /**
+     * JIRA: POOL-287
+     * 
+     * Verify that when an attempt is made to borrow an instance from the pool
+     * while the evictor is visiting it, there is no capacity leak.
+     * 
+     * Test creates the scenario described in POOL-287.
+     */
+    @Test
+    public void testReturnToHead() throws Exception {
+        final SimpleFactory<String> factory = new SimpleFactory<String>();
+        factory.setValidateLatency(100);
+        factory.setValid(true);  // Validation always succeeds
+        final GenericKeyedObjectPool<String, String> pool = new GenericKeyedObjectPool<String, String>(factory);
+        pool.setMaxWaitMillis(1000);
+        pool.setTestWhileIdle(true);
+        pool.setMaxTotalPerKey(2);
+        pool.setNumTestsPerEvictionRun(1);
+        pool.setTimeBetweenEvictionRunsMillis(500);
+        
+        // Load pool with two objects
+        pool.addObject("one");  // call this o1
+        pool.addObject("one");  // call this o2
+        // Default is LIFO, so "one" pool is now [o2, o1] in offer order.
+        // Evictor will visit in oldest-to-youngest order, so o1 then o2
+        
+        Thread.sleep(800); // Wait for first eviction run to complete
+        
+        // At this point, one eviction run should have completed, visiting o1
+        // and eviction cursor should be pointed at o2, which is the next offered instance
+        Thread.sleep(250);         // Wait for evictor to start
+        final String o1 = pool.borrowObject("one");  // o2 is under eviction, so this will return o1
+        final String o2 = pool.borrowObject("one");  // Once validation completes, o2 should be offered
+        pool.returnObject("one", o1);
+        pool.returnObject("one", o2);
+        pool.close();
+    }
+    
+    /**
+     * Verifies that when a factory's makeObject produces instances that are not
+     * discernible by equals, the pool can handle them.  
+     * 
+     * JIRA: POOL-283
+     */
+    @Test
+    public void testEqualsIndiscernible() throws Exception {
+        final HashSetFactory factory = new HashSetFactory();
+        final GenericKeyedObjectPool<String,HashSet<String>> pool =
+                new GenericKeyedObjectPool<String, HashSet<String>>(
+                        factory, new GenericKeyedObjectPoolConfig());
+        final HashSet<String> s1 = pool.borrowObject("a");
+        final HashSet<String> s2 = pool.borrowObject("a");
+        pool.returnObject("a", s1);
+        pool.returnObject("a", s2);
+        pool.close();
+    }
+    
+    /**
+     * Verifies that when a borrowed object is mutated in a way that does not
+     * preserve equality and hashcode, the pool can recognized it on return.
+     * 
+     * JIRA: POOL-284
+     */
+    @Test
+    public void testMutable() throws Exception {
+        final HashSetFactory factory = new HashSetFactory();
+        final GenericKeyedObjectPool<String,HashSet<String>> pool =
+                new GenericKeyedObjectPool<String, HashSet<String>>(
+                        factory, new GenericKeyedObjectPoolConfig());
+        final HashSet<String> s1 = pool.borrowObject("a");
+        final HashSet<String> s2 = pool.borrowObject("a");
+        s1.add("One");
+        s2.add("One");
+        pool.returnObject("a", s1);
+        pool.returnObject("a", s2);
+        pool.close();
+    }
+    
+    /**
+     * Verifies that returning an object twice (without borrow in between) causes ISE
+     * but does not re-validate or re-passivate the instance.
+     * 
+     * JIRA: POOL-285
+     */
+    @Test
+    public void testMultipleReturn() throws Exception {
+        final WaiterFactory<String> factory = new WaiterFactory<String>(0, 0, 0, 0, 0, 0);
+        final GenericKeyedObjectPool<String, Waiter> pool =
+                new GenericKeyedObjectPool<String, Waiter>(factory);
+        pool.setTestOnReturn(true);
+        Waiter waiter = pool.borrowObject("a");
+        pool.returnObject("a",waiter);
+        Assert.assertEquals(1, waiter.getValidationCount());
+        Assert.assertEquals(1, waiter.getPassivationCount());
+        try {
+            pool.returnObject("a",waiter);
+            fail("Expecting IllegalStateException from multiple return");
+        } catch (IllegalStateException ex) {
+            // Exception is expected, now check no repeat validation/passivation
+            Assert.assertEquals(1, waiter.getValidationCount());
+            Assert.assertEquals(1, waiter.getPassivationCount());
+        }
+    }
 
     private static class DummyFactory
             extends BaseKeyedPooledObjectFactory<Object,Object> {
@@ -2148,6 +2258,23 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
             return new DefaultPooledObject<Object>(value);
         }
     }
+    
+    /** 
+     * Factory that creates HashSets.  Note that this means
+     *  0) All instances are initially equal (not discernible by equals)
+     *  1) Instances are mutable and mutation can cause change in identity / hashcode.
+     */
+    private static final class HashSetFactory 
+            extends BaseKeyedPooledObjectFactory<String, HashSet<String>> {
+        @Override
+        public HashSet<String> create(String key) throws Exception {
+            return new HashSet<String>();
+        }
+        @Override
+        public PooledObject<HashSet<String>> wrap(HashSet<String> value) {
+            return new DefaultPooledObject<HashSet<String>>(value);
+        }
+    }
 
     private static class SimplePerKeyFactory
             extends BaseKeyedPooledObjectFactory<Object,Object> {
