@@ -26,10 +26,13 @@ import static org.junit.Assert.fail;
 
 import java.lang.management.ManagementFactory;
 import java.util.ArrayList;
+import java.util.HashSet;
 import java.util.List;
 import java.util.NoSuchElementException;
 import java.util.Random;
 import java.util.Set;
+import java.util.Timer;
+import java.util.TimerTask;
 import java.util.concurrent.atomic.AtomicInteger;
 
 import javax.management.MBeanServer;
@@ -44,13 +47,15 @@ import org.apache.commons.pool2.SwallowedExceptionListener;
 import org.apache.commons.pool2.TestBaseObjectPool;
 import org.apache.commons.pool2.VisitTracker;
 import org.apache.commons.pool2.VisitTrackerFactory;
+import org.apache.commons.pool2.Waiter;
+import org.apache.commons.pool2.WaiterFactory;
 import org.junit.After;
 import org.junit.Assert;
 import org.junit.Before;
 import org.junit.Test;
 
 /**
- * @version $Revision: 1569016 $
+ * @version $Revision: 1679560 $
  */
 public class TestGenericObjectPool extends TestBaseObjectPool {
 
@@ -289,7 +294,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         pool.borrowObject(); // numActive = 1, numIdle = 0
         // Create a test thread that will run once and try a borrow after
         // 150ms fixed delay
-        TestThread borrower = new TestThread(pool, 1, 150, false);
+        TestThread<String> borrower = new TestThread<String>(pool, 1, 150, false);
         Thread borrowerThread = new Thread(borrower);
         // Set evictor to run in 100 ms - will create idle instance
         pool.setTimeBetweenEvictionRunsMillis(100);
@@ -677,6 +682,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
     }
 
     @Test(timeout=60000)
+    @SuppressWarnings("rawtypes")
     public void testMaxTotalUnderLoad() {
         // Config
         int numThreads = 199; // And main thread makes a round 200.
@@ -695,7 +701,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             // Factor of 2 on iterations so main thread does work whilst other
             // threads are running. Factor of 2 on delay so average delay for
             // other threads == actual delay for main thread
-            threads[i] = new TestThread(pool, numIter * 2, delay * 2);
+            threads[i] = new TestThread<String>(pool, numIter * 2, delay * 2);
             Thread t = new Thread(threads[i]);
             t.start();
         }
@@ -753,6 +759,45 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         }
     }
 
+    /**
+     * This is the test case for POOL-263. It is disabled since it will always
+     * pass without artificial delay being injected into GOP.returnObject() and
+     * a way to this this hasn't currently been found that doesn't involve
+     * polluting the GOP implementation. The artificial delay needs to be
+     * inserted just before the final call to isLifo() in the returnObject()
+     * method.
+     */
+    //@Test(timeout=60000)
+    public void testReturnObject() throws Exception {
+
+        pool.setMaxTotal(1);
+        pool.setMaxIdle(-1);
+        String active = pool.borrowObject();
+
+        assertEquals(1, pool.getNumActive());
+        assertEquals(0, pool.getNumIdle());
+
+        Thread t = new Thread() {
+
+            @Override
+            public void run() {
+                pool.close();
+            }
+
+        };
+        t.start();
+
+        pool.returnObject(active);
+
+        // Wait for the close() thread to complete
+        while (t.isAlive()) {
+            Thread.sleep(50);
+        }
+
+        assertEquals(0, pool.getNumIdle());
+    }
+
+
     @Test(timeout=60000)
     public void testSettersAndGetters() throws Exception {
         {
@@ -988,7 +1033,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
 
     public static class TestEvictionPolicy<T> implements EvictionPolicy<T> {
 
-        private AtomicInteger callCount = new AtomicInteger(0);
+        private final AtomicInteger callCount = new AtomicInteger(0);
 
         @Override
         public boolean evict(EvictionConfig config, PooledObject<T> underTest,
@@ -1316,10 +1361,13 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
      * <iterations> borrow-return cycles with random delay times <= delay
      * in between.
      */
-    private void runTestThreads(int numThreads, int iterations, int delay) {
+    @SuppressWarnings({
+        "rawtypes", "unchecked"
+    })
+    private void runTestThreads(int numThreads, int iterations, int delay, GenericObjectPool testPool) {
         TestThread[] threads = new TestThread[numThreads];
         for(int i=0;i<numThreads;i++) {
-            threads[i] = new TestThread(pool,iterations,delay);
+            threads[i] = new TestThread<String>(testPool,iterations,delay);
             Thread t = new Thread(threads[i]);
             t.start();
         }
@@ -1342,7 +1390,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         pool.setMaxTotal(15);
         pool.setMaxIdle(15);
         pool.setMaxWaitMillis(1000L);
-        runTestThreads(20, 100, 50);
+        runTestThreads(20, 100, 50, pool);
     }
 
     /**
@@ -1363,7 +1411,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         pool.setMaxIdle(-1);
         pool.setTestOnReturn(true);
         pool.setMaxWaitMillis(1000L);
-        runTestThreads(5, 10, 50);
+        runTestThreads(5, 10, 50, pool);
     }
 
     @Test(timeout=60000)
@@ -1393,6 +1441,26 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         }
     }
 
+    /**
+     * Verifies that concurrent threads never "share" instances
+     */
+    @Test
+    public void testNoInstanceOverlap() {
+        final int maxTotal = 5;
+        final int numThreads = 100;
+        final int delay = 1;
+        final int iterations = 1000;
+        AtomicIntegerFactory factory = new AtomicIntegerFactory();
+        GenericObjectPool<AtomicInteger> pool = new GenericObjectPool<AtomicInteger>(factory);
+        pool.setMaxTotal(maxTotal);
+        pool.setMaxIdle(maxTotal);
+        pool.setTestOnBorrow(true);
+        pool.setBlockWhenExhausted(true);
+        pool.setMaxWaitMillis(-1);
+        runTestThreads(numThreads, iterations, delay, pool);
+        Assert.assertEquals(0, pool.getDestroyedByBorrowValidationCount());
+    }
+
     /**
      * POOL-189
      *
@@ -1425,7 +1493,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
     }
 
     private class ConcurrentBorrowAndEvictThread extends Thread {
-        private boolean borrow;
+        private final boolean borrow;
         public String obj;
 
         public ConcurrentBorrowAndEvictThread(boolean borrow) {
@@ -1444,46 +1512,66 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         }
     }
 
-    static class TestThread implements Runnable {
-        private final java.util.Random _random = new java.util.Random();
+    static class TestThread<T> implements Runnable {
 
-        // Thread config items
-        private final ObjectPool<String> _pool;
+        /** source of random delay times */
+        private final java.util.Random _random;
+
+        /** pool to borrow from */
+        private final ObjectPool<T> _pool;
+
+        /** number of borrow attempts */
         private final int _iter;
-        private final int _delay;
+
+        /** delay before each borrow attempt */
+        private final int _startDelay;
+
+        /** time to hold each borrowed object before returning it */
+        private final int _holdTime;
+
+        /** whether or not start and hold time are randomly generated */
         private final boolean _randomDelay;
+
+        /** object expected to be borrowed (fail otherwise) */
         private final Object _expectedObject;
 
         private volatile boolean _complete = false;
         private volatile boolean _failed = false;
         private volatile Throwable _error;
 
-        public TestThread(ObjectPool<String> pool) {
+        public TestThread(ObjectPool<T> pool) {
             this(pool, 100, 50, true, null);
         }
 
-        public TestThread(ObjectPool<String> pool, int iter) {
+        public TestThread(ObjectPool<T> pool, int iter) {
             this(pool, iter, 50, true, null);
         }
 
-        public TestThread(ObjectPool<String> pool, int iter, int delay) {
+        public TestThread(ObjectPool<T> pool, int iter, int delay) {
             this(pool, iter, delay, true, null);
         }
 
-        public TestThread(ObjectPool<String> pool, int iter, int delay,
+        public TestThread(ObjectPool<T> pool, int iter, int delay,
                 boolean randomDelay) {
             this(pool, iter, delay, randomDelay, null);
         }
 
-        public TestThread(ObjectPool<String> pool, int iter, int delay,
+        public TestThread(ObjectPool<T> pool, int iter, int delay,
                 boolean randomDelay, Object obj) {
-            _pool = pool;
-            _iter = iter;
-            _delay = delay;
-            _randomDelay = randomDelay;
-            _expectedObject = obj;
+            this(pool, iter, delay, delay, randomDelay, obj);
         }
 
+        public TestThread(ObjectPool<T> pool, int iter, int startDelay,
+            int holdTime, boolean randomDelay, Object obj) {
+        _pool = pool;
+        _iter = iter;
+        _startDelay = startDelay;
+        _holdTime = holdTime;
+        _randomDelay = randomDelay;
+        _random = _randomDelay ? new Random() : null;
+        _expectedObject = obj;
+    }
+
         public boolean complete() {
             return _complete;
         }
@@ -1495,14 +1583,16 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         @Override
         public void run() {
             for(int i=0;i<_iter;i++) {
-                long delay =
-                    _randomDelay ? (long)_random.nextInt(_delay) : _delay;
+                long startDelay =
+                    _randomDelay ? (long)_random.nextInt(_startDelay) : _startDelay;
+                long holdTime =
+                    _randomDelay ? (long)_random.nextInt(_holdTime) : _holdTime;
                 try {
-                    Thread.sleep(delay);
+                    Thread.sleep(startDelay);
                 } catch(InterruptedException e) {
                     // ignored
                 }
-                String obj = null;
+                T obj = null;
                 try {
                     obj = _pool.borrowObject();
                 } catch(Exception e) {
@@ -1520,7 +1610,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
                 }
 
                 try {
-                    Thread.sleep(delay);
+                    Thread.sleep(holdTime);
                 } catch(InterruptedException e) {
                     // ignored
                 }
@@ -1717,7 +1807,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
                 hurl = exceptionOnActivate;
                 evenTest = evenValid;
                 oddTest = oddValid;
-                counter = validateCounter++;
+                counter = activationCounter++;
             }
             if (hurl) {
                 if (!(counter%2 == 0 ? evenTest : oddTest)) {
@@ -1736,6 +1826,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             }
         }
         int makeCounter = 0;
+        int activationCounter = 0;
         int validateCounter = 0;
         int activeCount = 0;
         boolean evenValid = true;
@@ -1781,6 +1872,101 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             }
         }
     }
+
+    protected static class AtomicIntegerFactory
+    extends BasePooledObjectFactory<AtomicInteger> {
+
+    private long activateLatency = 0;
+    private long passivateLatency = 0;
+    private long createLatency = 0;
+    private long destroyLatency = 0;
+    private long validateLatency = 0;
+
+    @Override
+    public AtomicInteger create() {
+        try {
+            Thread.sleep(createLatency);
+        } catch (InterruptedException ex) {}
+        return new AtomicInteger(0);
+    }
+
+    @Override
+    public PooledObject<AtomicInteger> wrap(AtomicInteger integer) {
+        return new DefaultPooledObject<AtomicInteger>(integer);
+    }
+
+    @Override
+    public void activateObject(PooledObject<AtomicInteger> p) {
+        p.getObject().incrementAndGet();
+        try {
+            Thread.sleep(activateLatency);
+        } catch (InterruptedException ex) {}
+    }
+
+    @Override
+    public void passivateObject(PooledObject<AtomicInteger> p) {
+        p.getObject().decrementAndGet();
+        try {
+            Thread.sleep(passivateLatency);
+        } catch (InterruptedException ex) {}
+    }
+
+    @Override
+    public boolean validateObject(PooledObject<AtomicInteger> instance) {
+        try {
+            Thread.sleep(validateLatency);
+        } catch (InterruptedException ex) {}
+        return instance.getObject().intValue() == 1;
+    }
+
+    @Override
+    public void destroyObject(PooledObject<AtomicInteger> p) {
+        try {
+            Thread.sleep(destroyLatency);
+        } catch (InterruptedException ex) {}
+    }
+
+
+    /**
+     * @param activateLatency the activateLatency to set
+     */
+    public void setActivateLatency(long activateLatency) {
+        this.activateLatency = activateLatency;
+    }
+
+
+    /**
+     * @param passivateLatency the passivateLatency to set
+     */
+    public void setPassivateLatency(long passivateLatency) {
+        this.passivateLatency = passivateLatency;
+    }
+
+
+    /**
+     * @param createLatency the createLatency to set
+     */
+    public void setCreateLatency(long createLatency) {
+        this.createLatency = createLatency;
+    }
+
+
+    /**
+     * @param destroyLatency the destroyLatency to set
+     */
+    public void setDestroyLatency(long destroyLatency) {
+        this.destroyLatency = destroyLatency;
+    }
+
+
+    /**
+     * @param validateLatency the validateLatency to set
+     */
+    public void setValidateLatency(long validateLatency) {
+        this.validateLatency = validateLatency;
+    }
+}
+
     @Override
     protected boolean isLifo() {
         return true;
@@ -1796,31 +1982,48 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
      * enough margin for this to work correctly on most (all?) systems but be
      * aware of this if you see a failure of this test.
      */
+    @SuppressWarnings({
+        "rawtypes", "unchecked"
+    })
     @Test(timeout=60000)
-    public void testBorrowObjectFairness() {
+    public void testBorrowObjectFairness() throws Exception {
 
-        // Config
-        int numThreads = 30;
-        int maxTotal = 10;
+        int numThreads = 40;
+        int maxTotal = 40;
 
-        pool.setMaxTotal(maxTotal);
-        pool.setBlockWhenExhausted(true);
-        pool.setTimeBetweenEvictionRunsMillis(-1);
+        GenericObjectPoolConfig config = new GenericObjectPoolConfig();
+        config.setMaxTotal(maxTotal);
+        config.setMaxIdle(maxTotal);
+        config.setFairness(true);
+        config.setLifo(false);
 
-        // Start threads to borrow objects
+        pool = new GenericObjectPool(factory, config);
+
+        // Exhaust the pool
+        String[] objects = new String[maxTotal];
+        for (int i = 0; i < maxTotal; i++) {
+            objects[i] = pool.borrowObject();
+        }
+
+        // Start and park threads waiting to borrow objects
         TestThread[] threads = new TestThread[numThreads];
         for(int i=0;i<numThreads;i++) {
-            threads[i] = new TestThread(pool, 1, 2000, false, String.valueOf(i % maxTotal));
+            threads[i] = new TestThread(pool, 1, 0, 2000, false, String.valueOf(i % maxTotal));
             Thread t = new Thread(threads[i]);
             t.start();
             // Short delay to ensure threads start in correct order
             try {
-                Thread.sleep(50);
+                Thread.sleep(10);
             } catch (InterruptedException e) {
                 fail(e.toString());
             }
         }
 
+        // Return objects, other threads should get served in order
+        for (int i = 0; i < maxTotal; i++) {
+            pool.returnObject(objects[i]);
+        }
+
         // Wait for threads to finish
         for(int i=0;i<numThreads;i++) {
             while(!(threads[i]).complete()) {
@@ -1948,13 +2151,13 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             wtt[i] = new WaitingTestThread(pool,holdTime);
         }
         long origin = System.currentTimeMillis()-1000;
-        for(int i=0; i < wtt.length; i++){
-            wtt[i].start();
+        for (WaitingTestThread element : wtt) {
+            element.start();
         }
         int failed = 0;
-        for(int i=0; i < wtt.length; i++){
-            wtt[i].join();
-            if (wtt[i]._thrown != null){
+        for (WaitingTestThread element : wtt) {
+            element.join();
+            if (element._thrown != null){
                 failed++;
             }
         }
@@ -1966,8 +2169,7 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
                     " Threads: " + wtt.length +
                     " Failed: " + failed
                     );
-            for(int i=0; i < wtt.length; i++){
-                WaitingTestThread wt = wtt[i];
+            for (WaitingTestThread wt : wtt) {
                 System.out.println(
                         "Preborrow: " + (wt.preborrow-origin) +
                         " Postborrow: " + (wt.postborrow != 0 ? wt.postborrow-origin : -1) +
@@ -2153,6 +2355,130 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
         Assert.assertEquals(1, pool.getNumIdle());
     }
 
+    // POOL-259
+    @Test
+    public void testClientWaitStats() throws Exception {
+        SimpleFactory factory = new SimpleFactory();
+        // Give makeObject a little latency
+        factory.setMakeLatency(200);
+        final GenericObjectPool<String> pool = new GenericObjectPool<String>(
+                factory, new GenericObjectPoolConfig());
+        String s = pool.borrowObject();
+        // First borrow waits on create, so wait time should be at least 200 ms
+        // Allow 100ms error in clock times
+        Assert.assertTrue(pool.getMaxBorrowWaitTimeMillis() >= 100);
+        Assert.assertTrue(pool.getMeanBorrowWaitTimeMillis() >= 100);
+        pool.returnObject(s);
+        pool.borrowObject();
+        // Second borrow does not have to wait on create, average should be about 100
+        Assert.assertTrue(pool.getMaxBorrowWaitTimeMillis() > 100);
+        Assert.assertTrue(pool.getMeanBorrowWaitTimeMillis() < 200);
+        Assert.assertTrue(pool.getMeanBorrowWaitTimeMillis() > 20);
+    }
+
+    // POOL-276
+    @Test
+    public void testValidationOnCreateOnly() throws Exception {
+        pool.setMaxTotal(1);
+        pool.setTestOnCreate(true);
+        pool.setTestOnBorrow(false);
+        pool.setTestOnReturn(false);
+        pool.setTestWhileIdle(false);
+
+        final String o1 = pool.borrowObject();
+        Assert.assertEquals("0", o1);
+        Timer t = new Timer();
+        t.schedule(
+                new TimerTask() {
+                    @Override
+                    public void run() {
+                        pool.returnObject(o1);
+                    }
+                }, 3000);
+
+        String o2 = pool.borrowObject();
+        Assert.assertEquals("0", o2);
+
+        Assert.assertEquals(1, factory.validateCounter);
+    }
+    
+    /**
+     * Verifies that when a factory's makeObject produces instances that are not
+     * discernible by equals, the pool can handle them.  
+     * 
+     * JIRA: POOL-283
+     */
+    @Test
+    public void testEqualsIndiscernible() throws Exception {
+        final HashSetFactory factory = new HashSetFactory();
+        final GenericObjectPool<HashSet<String>> pool = new GenericObjectPool<HashSet<String>>(
+                factory, new GenericObjectPoolConfig());
+        final HashSet<String> s1 = pool.borrowObject();
+        final HashSet<String> s2 = pool.borrowObject();
+        pool.returnObject(s1);
+        pool.returnObject(s2);
+        pool.close();
+    }
+    
+    /**
+     * Verifies that when a borrowed object is mutated in a way that does not
+     * preserve equality and hashcode, the pool can recognized it on return.
+     * 
+     * JIRA: POOL-284
+     */
+    @Test
+    public void testMutable() throws Exception {
+        final HashSetFactory factory = new HashSetFactory();
+        final GenericObjectPool<HashSet<String>> pool = new GenericObjectPool<HashSet<String>>(
+                factory, new GenericObjectPoolConfig());
+        final HashSet<String> s1 = pool.borrowObject();
+        final HashSet<String> s2 = pool.borrowObject();
+        s1.add("One");
+        s2.add("One");
+        pool.returnObject(s1);
+        pool.returnObject(s2);
+        pool.close();
+    }
+    
+    /**
+     * Verifies that returning an object twice (without borrow in between) causes ISE
+     * but does not re-validate or re-passivate the instance.
+     * 
+     * JIRA: POOL-285
+     */
+    @Test
+    public void testMultipleReturn() throws Exception {
+        final WaiterFactory<String> factory = new WaiterFactory<String>(0, 0, 0, 0, 0, 0);
+        final GenericObjectPool<Waiter> pool = new GenericObjectPool<Waiter>(factory);
+        pool.setTestOnReturn(true);
+        Waiter waiter = pool.borrowObject();
+        pool.returnObject(waiter);
+        Assert.assertEquals(1, waiter.getValidationCount());
+        Assert.assertEquals(1, waiter.getPassivationCount());
+        try {
+            pool.returnObject(waiter);
+            fail("Expecting IllegalStateException from multiple return");
+        } catch (IllegalStateException ex) {
+            // Exception is expected, now check no repeat validation/passivation
+            Assert.assertEquals(1, waiter.getValidationCount());
+            Assert.assertEquals(1, waiter.getPassivationCount());
+        }
+    }
+    
+    public void testPreparePool() throws Exception {
+        pool.setMinIdle(1);
+        pool.setMaxTotal(1);
+        pool.preparePool();
+        Assert.assertEquals(1, pool.getNumIdle());
+        String obj = pool.borrowObject();
+        pool.preparePool();
+        Assert.assertEquals(0, pool.getNumIdle());
+        pool.setMinIdle(0);
+        pool.returnObject(obj);
+        pool.preparePool();
+        Assert.assertEquals(0, pool.getNumIdle());
+    }
+
     private static final class DummyFactory
             extends BasePooledObjectFactory<Object> {
         @Override
@@ -2164,7 +2490,23 @@ public class TestGenericObjectPool extends TestBaseObjectPool {
             return new DefaultPooledObject<Object>(value);
         }
     }
-
+    
+    /** 
+     * Factory that creates HashSets.  Note that this means
+     *  0) All instances are initially equal (not discernible by equals)
+     *  1) Instances are mutable and mutation can cause change in identity / hashcode.
+     */
+    private static final class HashSetFactory 
+            extends BasePooledObjectFactory<HashSet<String>> {
+        @Override
+        public HashSet<String> create() throws Exception {
+            return new HashSet<String>();
+        }
+        @Override
+        public PooledObject<HashSet<String>> wrap(HashSet<String> value) {
+            return new DefaultPooledObject<HashSet<String>>(value);
+        }
+    }
 
     private static class InvalidFactory
             extends BasePooledObjectFactory<Object> {
